# -*- coding: utf-8 -*-
"""
Created on Wed Dec 21 10:33:10 2022

@author: jalon
"""

from sys import stdout
import numpy as np
import pandas as pd
from scipy import stats
from scipy.stats import pearsonr
import matplotlib.pyplot as plt
from sklearn.cross_decomposition import PLSRegression
from sklearn.model_selection import cross_val_predict
from sklearn.metrics import mean_squared_error, r2_score, mean_absolute_error
from sklearn import preprocessing
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import FunctionTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import explained_variance_score
from scipy.signal import find_peaks

 
def openCsv(url):
    """
    Opens the csv with path specified in url
    url: path to the csv with data
    data: dataFrame with the data in the file pointed by url
    dataTop: names of the columns of data
    t: temporal marks of the measurements
    moxCol: names of the columns correspondant to each sensor
    R: dataFrame with the values of the resistance of each MOx
    """
    pltSize= 5000
    data = pd.read_csv(url)
    # Get features
    dataTop = data.columns
    t = data[dataTop[0]].values
    moxCol = dataTop[6:]
    R = data[moxCol]
    
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()
    ax3 = ax1.twinx()
    for col in moxCol:    
        ax1.plot(t[0:pltSize], R[0:pltSize][col])
        ax1.set_xlabel(dataTop[0])
        ax1.set_ylabel('Resistance (M$\Omega$)')
        ax2.plot(t[0:pltSize],data[0:pltSize][dataTop[2]],c='blue')
        ax2.set_ylabel(dataTop[2],c='blue')
        ax3.spines.right.set_position(("axes", 1.05))
        ax3.plot(t[0:pltSize],data[0:pltSize][dataTop[1]],c='red')
        ax3.set_ylabel(dataTop[1],c='red')
    
    return(data,dataTop,t,moxCol,R)


def featSel3 (data, Nsamp, window, hvCycles):
    
    dataTop = data.columns
    moxCol = dataTop[6:]
    rh = data[dataTop[2]].values
    hv = data[dataTop[5]].values
    co = data[dataTop[1]].values
    t = data[dataTop[0]].values
    
    #median and moving avergae filter
    medMOX = 1/((10**3)*data[moxCol].rolling(window).median())
    filtMox = medMOX.rolling(window).mean()
    
    devHV = np.diff(hv)  
    peaksHV,_ = find_peaks(devHV,height=0.1) 
    peaksHV=np.concatenate(([0],peaksHV-1, [len(hv)]))
    intR = pd.DataFrame()
    rhF = []
    coF = []
    #Selection of the MOX signals after every heating sequence
    #The final condition of [CO] and rh(%) of each heating sequence is stored
    for i in range(0,len(peaksHV)-1,1):
        
        tx = t[peaksHV[i]:peaksHV[i+1]-1]
        x = filtMox[:][peaksHV[i]:peaksHV[i+1]-1]
        tint = np.linspace(tx[0],tx[-1],Nsamp)
        oneIntR = pd.DataFrame()
        for col in moxCol:
            xd = x[col].values
            xI = np.interp(tint,tx,xd)
            df = pd.DataFrame({col:xI})
            oneIntR = pd.concat([oneIntR,df],axis=1)
        
        rhF.append(rh[peaksHV[i+1]-1])
        coF.append(co[peaksHV[i+1]-1])
        intR = pd.concat([intR,oneIntR],axis=0).reset_index(drop=True)
        comp = 100*(i+1)/(len(peaksHV)-1)
        stdout.write("\r%d%% Cycle Selection completed" % comp)
        stdout.flush() 
        
    stdout.write("\n")
    X= intR.fillna(0)
    #Getting and averaging the last 5 cycles of the [CO] condition
    XF = pd.DataFrame()
    coAVG = []
    rhAVG = []
    i=0
    for col in moxCol:
        a =X[col].values
        b=pd.DataFrame(a.reshape((int((len(X))/Nsamp)),Nsamp))
        oneMox=pd.DataFrame()
        for j in range(1,int(np.floor(2*len(b)/hvCycles))):
            aux3 = b.iloc[j*20-4:j*20-1].mean(axis=0)
            coAVG.append(np.mean(coF[j*20-4:j*20-1]))
            rhAVG.append(np.mean(rhF[j*20-4:j*20-1]))
            oneMox = pd.concat((oneMox,aux3.T),axis=1)
        
        XF = pd.concat((XF,oneMox),axis=1)
        comp = 100*(i+1)/(len(moxCol))
        i+=1
        stdout.write("\r%d%% Averaging completed" % comp)
        stdout.flush() 
    stdout.write("\n")
    XF.columns=range(0,XF.shape[1])
    return(XF,rhAVG,coAVG)
       
def findH(dataSel,RHavg,COavg,RHmax,RHmin,Nsamp):
    """
    Finds in a dataFrame the samples adquired at certain Humidity
    dataSel: dataFrame with samples after pre-selection
    RHavg: Humidity at wich the samples were pre-selected
    COavg: CO concentration at wich the samples were pre-selected
    RHmax: maximum relative to be analyze
    RHmin: minimum relative to be analyze
    Nsamp: number of features per MOx sensor
    aa: returned dataFrame with the samples when RH = [RHmin,RHmax]
    co: concentrations when RH = [RHmin,RHmax]
    rh: relative humidities in [RHmin,RHmax] range
    """
    rh1 = RHavg[0:2*Nsamp]
    idxj = np.where((np.array(rh1)<=RHmax)&(np.array(rh1)>=RHmin))
    idx=np.array([])
    for j in range(0,14):
        idx=np.concatenate((idx,idxj[0]+j*(2*Nsamp+1)))
    aa=dataSel[idx]
    co=np.array(COavg)[idx.astype(int)]
    rh=np.array(RHavg)[idx.astype(int)]    
    return(aa.T,co,rh)

def log_transform(x):
    print(x)
    return np.log10(x)
        
def logTrans(X):
    """
    Example of custom transformation with sklearn package.
    FunctionTransformater calls a previous defined function
    and transforms the dataFrame.
    X: dataFrame with relevant data
    XF transformed dataFrame
    """
    transformer = FunctionTransformer(log_transform)
    XF = transformer.transform(X)
    return (XF)

def scaling(X):
    """
    Mean-centering using StandardScaler from sklearn package
    X: dataFrame with relevant features
    XF: Men-centered dataFrame
    """
    scaler =StandardScaler(with_std=False).fit(X)
    XF = scaler.transform(X)
    return(XF)

def senSeparation(X):
    """
    Separetes the dataset in half
    X: dataFrame with every MOX sensor
    figSel: selected FIGARO sensors
    fisSel: selected FIS sensors
    """
    figSel=X.iloc[range(0,int(len(X)/2))]
    fisSel=X.iloc[range(int(len(X)/2),len(X))]
    return(figSel, fisSel)
  
def create_plots(true, pred):
    """
    Create plots to visualize the performance of a model's prediction
    Parameters:
        true : numpy.ndarray : The true values of the target column
        pred : numpy.ndarray : The predicted values of the target column
    """
    plt.figure(figsize=(16, 4))
    plt.subplot(1, 3, 1)
    plt.scatter(true, pred)
    plt.plot([0,np.max(pred)],[0,np.max(true)], "--r")
    plt.xlabel("Observed")
    plt.ylabel("Predicted")
    
    plt.subplot(1, 3, 2)
    pd.Series(true).hist(bins=10)
    pd.Series(pred).hist(bins=10)
    
    plt.subplot(1, 3, 3)
    pd.Series(true).plot()
    pd.Series(pred).plot()
    plt.legend(["observations", "predictions"])
    plt.show()
    
def optimise_pls_cv(X, y, n_comp, plot_components=True):
    
    """
    Optimise the number of latent variables of the PLS model
    and provides information of the optimized model 
    in the inner calibration.
    X: dataFrame with the features
    y: observed variable
    n_comp: maximum number of components to check
    plot_components: True if you want to see the evolution of RMSE
    pls_opt: optimized model
    y_c: predicted points during calibration
    y_cv: prediceted points during cross_validation
    """
    mse = []
    expVar = []
    component = np.arange(1, n_comp)
    
    for i in component:
        pls = PLSRegression(n_components=i)
 
        # Cross-validation
        y_cv = cross_val_predict(pls, X, y, cv=10)
 
        mse.append(np.sqrt(mean_squared_error(y, y_cv)))
        expVar.append(explained_variance_score(y,y_cv))
 
        comp = 100*(i+1)/n_comp
        # Trick to update status on the same line
        stdout.write("\r%d%% completed" % comp)
        stdout.flush()
    stdout.write("\n")
 
    # Calculate and print the position of minimum in MSE
    msemin = np.argmin(mse)
    print("Suggested number of components: ", msemin+1)
    stdout.write("\n")
 
    if plot_components is True:
            fig, ax1 = plt.subplots()
            ax2 = ax1.twinx()
            ax1.plot(component, np.array(mse), '-v', color = 'blue', mfc='blue')
            ax1.plot(component[msemin], np.array(mse)[msemin], 'P', ms=10, mfc='red')
            ax1.set_xlabel('Number of PLS components')
            ax1.set_ylabel('RMSE')
            plt.title('PLS')
            ax1.set_xlim(left=-1)
            ax2.plot(component, np.array(expVar)*100, '-v', color = 'orange', mfc='blue')
            ax2.plot(component[msemin], np.array(expVar)[msemin]*100, 'P', ms=10, mfc='red')
            ax2.set_ylabel('Explained Variance (%)')
 
 
 
    # Define PLS object with optimal number of components
    pls_opt = PLSRegression(n_components=msemin+1)
 
    # Fit to the entire dataset
    pls_opt.fit(X, y)
    y_c = pls_opt.predict(X)
 
    # Cross-validation
    y_cv = cross_val_predict(pls_opt, X, y, cv=10)
 
    #Calculate scores for calibration and cross-validation
    score_c = r2_score(y, y_c)
    score_cv = r2_score(y, y_cv)
 
    # Calculate mean squared error for calibration and cross validation
    mse_c = np.sqrt(mean_squared_error(y, y_c))
    mse_cv = np.sqrt(mean_squared_error(y, y_cv))
    # Calculate mean absolute error for calibration and cross validation    
    mae_c = mean_absolute_error(y, y_c)
    mae_cv = mean_absolute_error(y, y_cv)
    
    expVar_score = explained_variance_score(y,y_cv)*100
    
    
    print('R2 train: %5.3f'  % score_c)
    print('R2 internal CV: %5.3f'  % score_cv)
    print('RMSE train: %5.3f' % mse_c)
    print('RMSE internal CV: %5.3f' % mse_cv)
    print('MAE train: %5.3f' % mae_c)
    print('MAE internal CV: %5.3f' % mae_cv)
    print('Explained Variance internal CV: %5.3f' % expVar_score)
  
    return (pls_opt, y_c[:,0], y_cv[:,0], msemin+1)













    